import JsonDF.utils.Dict.Dict as d
import JsonDF.utils.List.List as l

class Data:
    def __init__(self, prefix='root', data=dict()) -> None:
        self.data = data
        self.prefix = prefix
        self.rows = {}
        self.rows_flatten = {}
    
    def childs(self):
        if type(self.data) == dict: 
            if len(self.data.keys()) >= 1:
                self.childsType(list(self.data.keys()))
        if type(self.data) == list:
            if len(self.data) >= 1:
                self.data = self.data[0]
                self.childsType(self.data)
    
    def childsType(self, keys):
        for key in keys:
            if type(self.data[key]) == list or type(self.data[key]) == dict:
                self.rows[f"{self.prefix}_{key}"] = self.child(data=self.data[key], prefix=f"{self.prefix}_{key}")
            else:
                self.rows[f"{self.prefix}_{key}"] = self.data[key]
    
    def tabelize(self, rows):
        for key, value in rows.items():
            if type(value) == dict:
                yield from self.tabelize(value)
            else:
                yield(key, value)
    
    def flatten(self):
        for key, value in self.tabelize(self.rows):
            self.rows_flatten[key] = value


    def child(self, data, prefix):
        if type(data) == list:
            return l.List(data=data, prefix=prefix).rows
        elif type(data) == dict:
            return d.Dict(data=data, prefix=prefix).rows


from ..Json import Json
import inspect

class Serializer:
    def __init__(self, object:object):
        self.object = object
    
    def get_name(self):
        return self.object.__name__
    
    def get_object_params(self):
        params = Json('params', {}).objectiy()
        attributes = inspect.getmembers(self.object, lambda a:not(inspect.isroutine(a)))
        for a in attributes:
            if not (a[0].startswith('__') and a[0].endswith('__')):
                params.insert(f"{a[0]}", a)
        return params
    
    def get_methods(self):
        methods = Json('methods', {}).objectiy()
        for key, value in self.object.__dict__.items():
            if callable(self.object.__dict__[key]):
                method_name = key
                method_params = self.get_method_params(value)
                method_code = self.get_code(value)
                methods.insert(f"{self.get_name()}.{method_name}", Json(f"{self.get_name()}.{method_name}", {f"{self.get_name()}.{method_name}": { 'params' : method_params, 'code': method_code }}).objectiy())
        return methods


    def get_method_params(self, method:classmethod) -> set:
        # print(method.__code__.co_varnames[:method.__code__.co_argcount])
        return (inspect.getfullargspec(method).args, inspect.getfullargspec(method).defaults)
    
    def get_code(self, method:classmethod) -> str:
        return inspect.getsource(method)
    
    def Serialize(self):
        return Json(self.get_name(), {'name': self.get_name(),
                                      'params': self.get_object_params(),
                                      'methods': self.get_methods()}).objectiy()


class Json:
    def __init__(self, name, json={}, depth=0):
        self.json = json
        self.json_name = name
        self.__depth__ = depth
        self.__total_depth__ = 0
        self.__depth_in__ = []
    
    def objectiy(self):
        keys = self.json.keys()
        for key in keys:
            value = self.__value(self.json[key], key)
            self.__setattr__(key, value)
        return self
    
    def insert(self, name, value):
        self.__setattr__(name, value)
        return self
    
    def delete(self, name):
        del self.__dict__[name]
        return self
    
    def dump(self, name):
        valueType = type(self.__getattribute__(name))
        if valueType == Json:
            self.__total_depth__ += 1
            self.__depth_in__.append(self.name)
            self.__setattr__(name, Json(0, {}, self.__depth__+1))
        elif valueType == list:
            self.__setattr__(name, [])
        elif valueType == dict:
            self.__setattr__(name, {})
        elif valueType == str:
            self.__setattr__(name, '')
        else:
            self.__setattr__(name, 0)
        return self
    
    def find(self, name, report=False):
        try:
            value = self.__getattribute__(name)
            return (value, True) if report else value
        except:
            return ((self.__depth__, f"No results for {name} in {self.json_name} at depth {self.__depth__}"), False) if report else f"No results for {name} in {self.json_name} at depth {self.__depth__}"
    def find_all(self, name, reports=True):
        query = {}
        result = self.find(name, True)
        if result[1]:
            query[(self.__depth__, name, self.json_name)] = self.find(name, True)[0]
            print(query)
        else:
            if reports: print(result[0][1])

        if self.depth_check(self):
            for depth_point in self.__depth_in__:
                if type(self.__getattribute__(depth_point)) == list:
                    self.__getattribute__(depth_point)[0].find_all(name, reports)
                else:
                    self.__getattribute__(depth_point).find_all(name, reports)
        
        return query
    
    def depth_check(self, json:object):
        if json.__total_depth__ > 0:
            return True
        return False

    def __value(self, value, key):
        if type(value) == dict or type(value) == list:
            if type(value) == list:
                return [self.process(value, key)]
            else:
                self.__total_depth__ += 1
                self.__depth_in__.append(key)
                return Json(key, value, self.__depth__+1).objectiy()
        else:
            return value

    def process(self, value, key):
        if value == None: return []
        for val in value:
            if type(val) == dict:
                self.__total_depth__ += 1
                self.__depth_in__.append(key)
                return Json(key, val, self.__depth__+1).objectiy()
    
    def show(self):
        attrs = {}
        for attr in vars(self):
            if attr.startswith('__') or callable(attr) or attr == 'json' or attr == 'json_name':
                continue
            attrs[attr] = getattr(self, attr)
        return attrs

    def __repr__(self):
        return self.show().__str__()


import JsonDF.utils.Dict.Dict as d

class List:
    def __init__(self, prefix, data):
        self.data = data
        self.prefix = prefix
        self.rows = {}
        self.childs = self.process()

    
    def process(self):
        if len(self.data) > 1:
            self.data = self.data[0]
            self.childs(self.data)
        elif len(self.data) == 1:
            if type(self.data[0]) == dict:
                self.data = self.data[0]
                return self.childs(list(self.data.keys()))
            else:
                return self.data[0]
    
    def childs(self, keys):
        for key in keys:
            if type(self.data[key]) == list or type(self.data[key]) == dict:
                self.rows[f"{self.prefix}_{key}"] = self.childType(data=self.data[key], prefix=f"{self.prefix}_{key}")
            else:
                self.rows[f"{self.prefix}_{key}"] = self.data[key]
    
    def childType(self, data, prefix):
        if type(data) == list:
            return List(data=data, prefix=prefix).rows
        elif type(data) == dict:
            return d.Dict(data=data, prefix=prefix).rows
